
#ifndef PLS_ROOT_MASTER_TASK_H
#define PLS_ROOT_MASTER_TASK_H

#include <mutex>

#include "abstract_task.h"
#include "pls/internal/base/spin_lock.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            template<typename Function>
            class root_task : public abstract_task {
                Function function_;
                bool finished_;

                // Improvement: Remove lock and replace by atomic variable (performance)
                base::spin_lock finished_lock_;
            public:
                explicit root_task(Function function):
                    abstract_task{0, id{0}},
                    function_{function},
                    finished_{false} {}

                bool finished() {
                    std::lock_guard<base::spin_lock> lock{finished_lock_};
                    return finished_;
                }

                void execute() override {
                    function_();
                    {
                        std::lock_guard<base::spin_lock> lock{finished_lock_};
                        finished_ = true;
                    }
                }

                bool internal_stealing(abstract_task* /*other_task*/) override {
                    return false;
                }

                bool split_task(base::spin_lock* /*lock*/) override {
                    return false;
                }
            };

            template<typename Function>
            class root_worker_task : public abstract_task {
                root_task<Function>* master_task_;

            public:
                explicit root_worker_task(root_task<Function>* master_task):
                        abstract_task{0, id{0}},
                        master_task_{master_task} {}

                void execute() override {
                    do {
                        steal_work();
                    } while (!master_task_->finished());
                }

                bool internal_stealing(abstract_task* /*other_task*/) override {
                    return false;
                }

                bool split_task(base::spin_lock* /*lock*/) override {
                    return false;
                }
            };
        }
    }
}

#endif //PLS_ROOT_MASTER_TASK_H
